/*
Implementation by the Keccak Team, namely, Guido Bertoni, Joan Daemen,
Michaël Peeters, Gilles Van Assche and Ronny Van Keer,
hereby denoted as "the implementer".

For more information, feedback or questions, please refer to our website:
https://keccak.team/

To the extent possible under law, the implementer has waived all copyright
and related or neighboring rights to the source code in this file.
http://creativecommons.org/publicdomain/zero/1.0/
*/

#include "crypto_aead.h"

#include <string.h>
#include "KetjeMn.h"

int crypto_aead_encrypt(
    unsigned char *c,unsigned long long *clen,
    const unsigned char *m,unsigned long long mlen,
    const unsigned char *ad,unsigned long long adlen,
    const unsigned char *nsec,
    const unsigned char *npub,
    const unsigned char *k
    )
{
    KetjeMn_Instance instance;

    KetjeMn_Initialize(&instance, k, 128, npub, 81*8);
    KetjeMn_FeedAssociatedData(&instance, ad, adlen);
    KetjeMn_WrapPlaintext(&instance, m, c, mlen);
    *clen = mlen;
    KetjeMn_GetTag(&instance, c+mlen, 16);
    *clen += 16;

    return 0;
}

int crypto_aead_decrypt(
    unsigned char *m,unsigned long long *mlen,
    unsigned char *nsec,
    const unsigned char *c,unsigned long long clen,
    const unsigned char *ad,unsigned long long adlen,
    const unsigned char *npub,
    const unsigned char *k
    )
{
    KetjeMn_Instance instance;
    unsigned char tag[16];

    if (clen < 16)
        return -1;

    KetjeMn_Initialize(&instance, k, 128, npub, 81*8);
    KetjeMn_FeedAssociatedData(&instance, ad, adlen);
    *mlen = clen-16;
    KetjeMn_UnwrapCiphertext(&instance, c, m, *mlen);
    KetjeMn_GetTag(&instance, tag, 16);
    if (memcmp(tag, c+(*mlen), 16) != 0) {
        memset(m, 0, *mlen);
        return -1;
    }
    else
        return 0;
}
